//=============================================================================
/** @file		GLWidget.cpp
 *
 * Implements CGLWidget.
 *
    Shader Maker - a cross-platform GLSL editor.
    Copyright (C) 2007-2008 Markus Kramer
    For details, see main.cpp or COPYING.

=============================================================================*/

#include <GL/glew.h>

#include "Config.hpp"
#include "GLWidget.hpp"
#include "Vector.hpp"
#include "Camera.hpp"

#include <QMouseEvent>

#include <iostream>

//=============================================================================
//	CGLWidget implementation
//=============================================================================

// construction
CGLWidget::CGLWidget(const QGLFormat& format, ICameraState* cs, QWidget* parent)
    : QGLWidget(format, parent)
    , m_fpsCounter      (0)
    , m_fpsValue        (0)
    , m_fpsLastPeriod   (0)
    , m_initSucceeded   (false)
    , m_cameraState     (cs)
{
    // make sure the GL context is active
    makeCurrent();

    setFocusPolicy(Qt::StrongFocus);
    setMouseTracking(true);
}


CGLWidget::~CGLWidget()
{
    makeCurrent();
}


bool CGLWidget::isSuccessfullyInitialized() const
{
    return m_initSucceeded;
}


/*
========================
getDriverInfoString

 fills the GL info widget with text
========================
*/
QString CGLWidget::getDriverInfoString() const
{
    // the context must be valied!
    if( !context()->isValid() )
        return QString( "<invalid OpenGL context>" );

    // read OpenGL infos
    QString text;
    text += tr( "Vendor:   %1\n" ).arg( (const char*)glGetString( GL_VENDOR ) );
    text += tr( "Renderer: %1\n" ).arg( (const char*)glGetString( GL_RENDERER ) );
    text += tr( "Version:  %1\n" ).arg( (const char*)glGetString( GL_VERSION ) );
    text += tr( "\n" );

    // print framebuffer format
    QGLFormat fmt = context()->format();
    text += tr( "Framebuffer Format:\n" );
    text += tr( " RGBA bits:    (%1,%2,%3,%4)\n" ).
                arg( fmt.redBufferSize() ).
                arg( fmt.greenBufferSize() ).
                arg( fmt.blueBufferSize() ).
                arg( fmt.alphaBufferSize() );
    text += tr( " Depth bits:   %1\n" ).arg( fmt.depthBufferSize() );
    text += tr( " Stencil bits: %1\n" ).arg( fmt.stencilBufferSize() );
    text += tr( "\n" );

    // shading language version
    if( fmt.openGLVersionFlags() & fmt.OpenGL_Version_2_0 )
    {
        QString version( (const char*)glGetString( GL_SHADING_LANGUAGE_VERSION ) );
        text += QString( "Shading Language Version: %1\n" ).arg( version );
    }

    // check for geometry shader
    bool hasGeometryShader = false;

    std::string extentions((const char*)glGetString(GL_EXTENSIONS));
    hasGeometryShader = (extentions.find("GL_EXT_geometry_shader4") != std::string::npos);

    if(hasGeometryShader)
    {
        text += tr( "GL_EXT_geometry_shader4 available\n" );
    }
    else
    {
        text += tr( "GL_EXT_geometry_shader4 not supported\n" );
    }

    text += tr( "\n" );

    GLint i;
#define PRINT_CONSTANT(Name) \
            glGetIntegerv( Name, &i ); \
            text += tr( #Name " = %1\n" ).arg( i );

	PRINT_CONSTANT( GL_MAX_TEXTURE_SIZE );
	PRINT_CONSTANT( GL_MAX_TEXTURE_UNITS );
	PRINT_CONSTANT( GL_MAX_VERTEX_ATTRIBS );
	PRINT_CONSTANT( GL_MAX_VERTEX_UNIFORM_COMPONENTS );
	PRINT_CONSTANT( GL_MAX_VARYING_FLOATS ); // alias for GL_MAX_VARYING_COMPONENTS_EXT
	PRINT_CONSTANT( GL_MAX_COMBINED_TEXTURE_IMAGE_UNITS );
	PRINT_CONSTANT( GL_MAX_VERTEX_TEXTURE_IMAGE_UNITS );
	PRINT_CONSTANT( GL_MAX_TEXTURE_IMAGE_UNITS );
	PRINT_CONSTANT( GL_MAX_TEXTURE_COORDS );
	PRINT_CONSTANT( GL_MAX_FRAGMENT_UNIFORM_COMPONENTS );

	if(hasGeometryShader)
	{
		PRINT_CONSTANT( GL_MAX_GEOMETRY_TEXTURE_IMAGE_UNITS_EXT );
		PRINT_CONSTANT( GL_MAX_GEOMETRY_VARYING_COMPONENTS_EXT );
		PRINT_CONSTANT( GL_MAX_VERTEX_VARYING_COMPONENTS_EXT );
		PRINT_CONSTANT( GL_MAX_VARYING_COMPONENTS_EXT );
		PRINT_CONSTANT( GL_MAX_GEOMETRY_UNIFORM_COMPONENTS_EXT) ;
		PRINT_CONSTANT( GL_MAX_GEOMETRY_OUTPUT_VERTICES_EXT );
		PRINT_CONSTANT( GL_MAX_GEOMETRY_TOTAL_OUTPUT_COMPONENTS_EXT );
	}

#undef PRINT_CONSTANT

    return text;
}


/*
========================
initializeGL

 - the GL context is current (see Qt spec...)
 - the GL_VERSION string will give us the info we need to know
========================
*/
void CGLWidget::initializeGL()
{
    try
    {
        initializeGlew();
    }
    catch (std::exception& e)
    {
        std::cerr << "Error: " << e.what() << std::endl;
        return;
    }

    m_initSucceeded = true;

    std::clog << "Status: Using OpenGL " << glGetString(GL_VERSION)
              << ", GLEW " << glewGetString(GLEW_VERSION)
              << std::endl;

    glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    glViewport(0, 0, width(), height());

    // start FPS timer
    m_fpsTimer.start();
    m_fpsLastPeriod = m_fpsTimer.elapsed() - 1000;

    // start periodic refesh timer
    startTimer(CONFIG_REFRESH_INTERVAL);
    m_initSucceeded = true;
}


/*
========================
paintGL
========================
*/
void CGLWidget::paintGL()
{
    if (!m_initSucceeded)
        return;

    glViewport(0, 0, width(), height());

    emit render();

    drawFPS();
}


/*
========================
resizeGL
========================
*/
void CGLWidget::resizeGL(int width, int height)
{
    m_viewportSize = QSize( width, height );
}


/*
========================
sizeHint
========================
*/
QSize CGLWidget::sizeHint() const
{
    return QSize(512, 1024);
}


/*
========================
minimumSizeHint
========================
*/
QSize CGLWidget::minimumSizeHint() const
{
    return QSize(128, 128);
}


void CGLWidget::initializeGlew() const
{
    GLenum err = glewInit();
    // throw error if glewInit failed, something is seriously wrong
    if (GLEW_OK != err)
    {
        std::string error(reinterpret_cast<const char*>(glewGetErrorString(err)));
        throw std::runtime_error("Failed to initialize GLEW; " + error);
    }

    //Check for OpenGL Driver 3.0

}


/*
========================
timerEvent
========================
*/
void CGLWidget::timerEvent(QTimerEvent* timer)
{
    updateGL();
}


/*
========================
mouseMoveEvent
========================
*/
void CGLWidget::mouseMoveEvent(QMouseEvent* event)
{
    int deltaX = event->x() - m_lastMousePosition.x();
    int deltaY = event->y() - m_lastMousePosition.y();

    int buttons = event->buttons();
    int controlFlag = ( event->modifiers() & Qt::MetaModifier ); // MacOS X

    //
    // back to front
    //
    if( controlFlag || ( buttons & Qt::RightButton ) )
    {
        vec3_t t;
        m_cameraState->getCameraTranslation( t );
        t.z += 0.02f * float(deltaY);
        m_cameraState->setCameraTranslation( t );
    }

    //
    // left/right up/down
    //
    else if ( buttons & Qt::MidButton )
    {
        vec3_t t;
        m_cameraState->getCameraTranslation( t );
        t.x += 0.01f * float(deltaX);
        t.y -= 0.01f * float(deltaY);
        m_cameraState->setCameraTranslation( t );
    }

    //
    // rotation
    //
    else if( buttons & Qt::LeftButton )
    {
        mat4_t m;
        m_cameraState->getCameraRotation( m );

        // use the matrix stack for this multiplication...
        glPushMatrix();
        glLoadIdentity();

        // apply new rotation after old rotation
        glRotatef( 0.5f * float(deltaX), 0.0f, 1.0f, 0.0f );
        glRotatef( 0.5f * float(deltaY), 1.0f, 0.0f, 0.0f );
        glMultMatrixf( m.toConstFloatPointer() );

        // state update...
        glGetFloatv( GL_MODELVIEW_MATRIX, m.toFloatPointer() );
        m_cameraState->setCameraRotation( m );
        glPopMatrix();
    }

    // save state and accept event.
    m_lastMousePosition = event->pos();
    event->accept();
}


/*
========================
keyPressEvent
========================
*/
void CGLWidget::keyPressEvent(QKeyEvent* event)
{

}


/*
========================
drawFPS

 This updates the FPS counter and displays current frame rate.
 If VSYNC is enabled, then we get very unprecise results.
 We must include SwapBuffers() in execution time, because most
 rendering work is done there.
========================
*/
void CGLWidget::drawFPS()
{
    // update value
    int now = m_fpsTimer.elapsed();
    if( now - m_fpsLastPeriod >= 1000 )
    {
        m_fpsValue = m_fpsCounter;
        m_fpsCounter = 0;
        m_fpsLastPeriod = now;
    }

    // count this frame
    m_fpsCounter++;

    // format text
    QString text = QString( "%1 FPS" ).arg( m_fpsValue );
    int length = fontMetrics().width( text );

    // setup state and render text
    glUseProgram( 0 );
    glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );
    glDisable( GL_LIGHTING );
    glDisable( GL_TEXTURE_2D );
    glColor3f( 1,1,1 );
    renderText( m_viewportSize.width() - length - 1, font().pointSize() + 1, text );
}
