//=============================================================================
/** @file		GLWidget.hpp
 *
 * Defines the OpenGL rendering widget.
 *
    Shader Maker - a cross-platform GLSL editor.
    Copyright (C) 2007-2008 Markus Kramer
    For details, see main.cpp or COPYING.

=============================================================================*/

#ifndef __CGLWIDGET_HPP_INCLUDED__
#define __CGLWIDGET_HPP_INCLUDED__

#include <QGLWidget>
#include <QTime>

// forward declarations
class ICameraState;


//=============================================================================
// CGLWidget
//=============================================================================

/**
 * OpenGL 2.0 rendering widget.
 * This widget is used to create and manage an OpenGL 2.0 context.
 * It is also responsible for capturing mouse and keyboard events and
 * for their processing. It interacts with a ICameraState object by
 * manipulating it's rotation and translation properties.
 * The widget must be shown to initialize the rendering context.
 * \n\n
 * The widet emits render() events in a regular time interval.
 */
class CGLWidget : public QGLWidget
{
    Q_OBJECT

public:

    /**
     * Constructs a GL widget.
     * @param format Display format for the OpenGL context.
     * @param cs ICameraState object to manipulate.
     */
    explicit CGLWidget(const QGLFormat& format, ICameraState* cs, QWidget* parent = nullptr);
    ~CGLWidget();


    /**
     * Check, wether the OpenGL context is active.
     * If the widget owns a valid OpenGL 2.0 rendering context,
     * this method returns True. Otherwise it returns False.
     * @return Wether a valid OpenGL 2.0 context is active.
     */
    bool isSuccessfullyInitialized() const;


    /**
     * Returns a driver info string containing.
     * The string is intended to be displayed to the user.
     * The OpenGL context must be initialized to use this call.
     */
    QString getDriverInfoString() const;

signals:

    /**
     * Periodic render event.
     * The user of this widget can be notified, that it is time to
     * redraw it's scene. When this signal is processed, the GL widget
     * has already set the current viewport parameters, so yout can
     * access them with glGetFloatv( GL_VIEWPORT, ... ).
     */
    void render();

protected:

    // QWidget inherited
    void initializeGL()                  override;
    void paintGL()                       override;
    void resizeGL(int width, int height) override;

    // QGLWidget inherited
    QSize sizeHint()        const override;
    QSize minimumSizeHint() const override;

private:
    void initializeGlew() const;

    // event handling
    void timerEvent(QTimerEvent* timer)     override;
    void mouseMoveEvent(QMouseEvent* event) override;
    void keyPressEvent(QKeyEvent* event)    override;

    // FPS
    void drawFPS();
    int m_fpsCounter;
    int m_fpsValue;
    int m_fpsLastPeriod;
    QTime m_fpsTimer;

    // caught on resizeGL()
    QSize m_viewportSize;

    // store last position to calc deltas
    QPoint m_lastMousePosition;

    // wether OpenGL 2.0 and shader functions are available.
    bool m_initSucceeded;

    // the widgets transforms input events to camera transformations
    ICameraState* m_cameraState;
};

#endif // __CGLWIDGET_HPP_INCLUDED__
